#!/usr/bin/env python

#import os, re, time, urlparse, urllib, urllib2
from GuiPlot import *
#from GuiUtils import _logerrc, _logwarn, _loginfo, natsorted, thousands, timeseries
#from cherrypy import expose, HTTPError, tree, response, engine, thread_data, log
#from cherrypy.lib.static import serve_file
#from GuiUtils import natsorted,thousands,timeseries
from cherrypy import HTTPError, engine
from threading import Thread, Lock
from copy import deepcopy, copy
#from calendar import timegm
import time, datetime
import logging
import random
import urllib,urllib2
import socket
import re
from xml.dom.minidom import parseString

class ProdmonWorkspace:
  """
  Main production monitoring workspace class.
  Handles the current front end state, context,
  recent plots etc.
  """
  sessiondef = {'prodmon.view':'summary'}
  maxrecent = 5
  class ProdmonView:
    def __init__(self,workspace):
      self.workspace = workspace
    def _sessionState(self, session):
      
      view = session['prodmon.view']
      
      keys = ["%r:%r" % (key.replace('prodmon.%s.'%(view),''),session[key]) for key in session.keys() if (key.startswith('prodmon.%s.'%(view)))]
      state_extra = self._stateHook(session)
      if state_extra:
        keys.extend(state_extra)
      keystr = ','.join(keys)
      return "{kind: '%s', version: 0, %s}" % (self.pluginname,keystr)
    def _graphLabel(self,session):
      return ''
    def _sessionView(self,session,*args,**kwargs):
      pass
    def _stateHook(self,session):
      return None
  
  class ProdmonSummary(ProdmonView):
    label="summary"
    pluginname="ProdmonSummary"
    name="Summary"
    viewdef={}
    def _stateHook(self,session):
      stats = self.workspace.source.getStatsSummary()
      keys = []#['top5site:%r' % self.workspace.source.summaryList('site'),'top5wf:%r' % self.workspace.source.summaryList('wf')]
      keys.extend(['%r:%r' % (k,v) for k,v in stats.iteritems()])
      return keys
  class ProdmonPlot(ProdmonView):
    label="plot"
    pluginname="ProdmonPlot"
    name="FreePlot"
    
    plotdef={
         'sortby':'site',         
         'what':'evts_read_any',
         'job_type':'Any',
         'graph':'bar',
         'regex':'null',
         'period':48,
         'use_period':'true',
         'group':'null',
         'filters':[],
         'image_width':800,
         'image_height':600,
         'image_dpi':72,
         'starttime':'',
         'endtime':'',
         'baseurl':'',
         'optstr':'',
         'title':''
         }
    viewdef={'prodmon.plot.saved':{},
             'prodmon.plot.open':[plotdef],
             'prodmon.plot.current':0
             }
    def _sessionState(self,session):
      
      thisplot = ', '.join(["%r:%r"%(k,v) for k,v in session['prodmon.plot.open'][session['prodmon.plot.current']].items()])
      savedlist = session['prodmon.plot.saved'].keys()
      openstruct = [{'index':i,'title':o['title'],'baseurl':o['baseurl'],'optstr':o['optstr']} for i,o in enumerate(session['prodmon.plot.open'])]
      keystr = "open: %r, saved: %s, current: %r, %s" % (openstruct,savedlist,session['prodmon.plot.current'],thisplot)
      return "{kind: '%s', version: 0, %s}" % (self.pluginname,keystr)
    def _sessionView(self,session,*args,**kwargs):
      current_plot = session['prodmon.plot.open'][session['prodmon.plot.current']]
      if 'new' in kwargs:
        current_plot = deepcopy(current_plot)
        session['prodmon.plot.open'].append(current_plot)
        session['prodmon.plot.current']=len(session['prodmon.plot.open'])-1
      for key,value in kwargs.items():
        value=urllib.unquote(value)
        if key in ('sortby','what','job_type','starttime','endtime','regex','graph','group','image_width','image_height','image_dpi','period','use_period','title'):
          if self.workspace._verify(key,value):
            current_plot[key]=value
      if 'filter_name' in kwargs and 'filter_value' in kwargs:
        name = urllib.unquote(kwargs['filter_name'])
        value = urllib.unquote(kwargs['filter_value'])
        if '|' in name and '|' in value:
          current_plot['filters'] = [[n,v] for n,v in zip(name.split('|'),value.split('|'))]
        else:
          current_plot['filters'] = [[name,value]]
          
      
      self.workspace._updateURLs(session)
  
  def __init__(self,gui,rank,category,name,*args):
    """
    Create a new ProdmonWorkspace instance. Note that we need to find ProdmonSource in the server
    source list, or we'll exit ingloriously. 
    
    TODO: good mechanism for finding sources
    TODO: SiteDB source integration
    TODO: ProdRequest integration
    TODO: Workspace base class
    
    """
    gui._addJSFragment("%s/gui-overview.js" % gui.contentpath)
    gui._addJSFragment("%s/gui-prodmon.js" % gui.contentpath)
    gui._addCSSFragment("%s/style-prodmon.css" % gui.contentpath)
    self.gui = gui
    self.name = name
    self.rank=rank
    self.category=category
    self.views=[ProdmonWorkspace.ProdmonSummary(self),ProdmonWorkspace.ProdmonPlot(self)]
    self.source = None
    for s in gui.sources:
      if s.plothook=='prodmon':
        self.source = s
    if self.source==None:
      raise Exception, "PW: Failed to find prodmon source. Fatal."
  
  
  def initialiseSession(self,session):
    """
    Called when session is created on the server. Populate the sessiondict from sessiondef.
    """
    for key,value in ProdmonWorkspace.sessiondef.iteritems():
      if key not in session:
        session[key]=deepcopy(value)
  
  def start(self,session,**kwargs):
    """
    Called when session lands on this workspace. Nothing to do here.
    """
    pass
  
  def sessionView(self,session,*args,**kwargs):
    """
    Called when a session navigates to a new view. Checks the view exists, then populates the sessiondict with the entries from the viewdef.
    """
    name = kwargs.get('name', None)
    if not isinstance(name, str):
      raise HTTPError(500, "Invalid name parameter")
    view = self._getView(name)
    for key,value in view.viewdef.iteritems():
      if key not in session:
        session[key]=deepcopy(value)
    view._sessionView(session,*args,**kwargs)
    session['prodmon.view'] = view.label
    self.gui._saveSession(session)
    return self._state(session)
  
  def _getView(self,label):
    """
    Helper function to resolve the current view object.
    """
    view = self.views[0]
    for v in self.views:
      if v.label == label:
        view = v
        break
    return view
  
  def _state(self,session):
    view = self._getView(session['prodmon.view'])
    self.gui._noResponseCaching()
    state = ("([{kind: 'AutoUpdate', interval: %d, stamp: %d},"
          + "{kind: 'OverviewHeader', workspace: '%s', workspaces: [%s],"
          + " view: '%s', views: [%s]},"
          + "%s])") \
          % (300,
           self.gui.stamp,
           self.name,
           ",".join("{title: '%s', label: '%s'}"
                % (x.name, x.name.lower()) for x in self.gui.workspaces),
                view.name,
                ",".join("{title: '%s', label: '%s'}"
                   % (x.name, x.label) for x in self.views),
                   view._sessionState(session)) 
    #print state
    return state
  
  def sessionState(self,session,*args,**kwargs):
    """
    Exposed method for _state.
    """  
    return self._state(session)
  
  def sessionResetWorkspace(self,session,*args,**kwargs):
    self.initialiseSession(session)
    self.gui._saveSession(session)
    return self._state(session) 
  
  def _verify(self,key,value):
    timepattern = re.compile('^[0-9]{4}-[0-9]{1,2}-[0-9]{1,2} [0-9]{2}:[0-9]{2}:[0-9]{2}$')
    verifier = {
      'sortby': lambda x: x in ProdmonSource.params['sortby'],
      'what': lambda x: x in ProdmonSource.params['what'],
      'job_type': lambda x: x in ProdmonSource.params['job_type'],
      'graph': lambda x: x in ProdmonSource.params['plot'],
      'use_period': lambda x: x in ('true','false'),
      'period': lambda x: x.isdigit(),
      'image_height': lambda x: x.isdigit(),
      'image_width': lambda x: x.isdigit(),
      'image_dpi': lambda x: x.isdigit(),
      'starttime': lambda x: timepattern.match(x),
      'endtime': lambda x: timepattern.match(x)
    }
    if key in verifier:
      if verifier[key](value):
        print "Verification succeeded for %s=%s" % (key,value)
        return value
      else:
        print "Verification failed for %s=%s" % (key,value)
        return ProdmonWorkspace.ProdmonPlot.plotdef[key]
    else:
      return value
      
  
  def _updateURLs(self,session):
    openplot =  session['prodmon.plot.open'][session['prodmon.plot.current']]
    baseurl = '/plotfairy/prodmon/%s/%s/%s/%s/' % (openplot['what'],openplot['sortby'],openplot['graph'],openplot['job_type']) #already verified that none of these should contain special characters
    
    opts = {}
    if not (openplot['regex']=='' or openplot['regex']=='null'):
      opts['regex'] = urllib.quote(openplot['regex'])
    if not (openplot['group']=='' or openplot['group']=='null'):
      opts['group'] = urllib.quote(openplot['group'])
    for f in openplot['filters']:
      opts['filter_%s' % f[0]] = urllib.quote(f[1])
    if openplot['use_period']=='true':
      opts['period']=openplot['period']
    else:
      opts['starttime']=urllib.quote(openplot['starttime'])
      opts['endtime']=urllib.quote(openplot['endtime'])
    opts['width']=openplot['image_width']
    opts['height']=openplot['image_height']
    opts['dpi']=openplot['image_dpi']
    
    optstr = urllib.urlencode(opts)
    openplot['baseurl']=baseurl
    openplot['optstr']=optstr
    
    if openplot['title']=='':
      openplot['title']='Unnamed Plot'
  
  def _formatTime(self,offset=0):
    return time.strftime('%Y-%m-%d %H:%M:%S',time.gmtime(time.time()-offset))
           
  def sessionGraph(self,session,*args,**kwargs):
    """
    Update one of the parameters (what, job_type, start/endtime, regex, graph type)
    Add the new parameter set to the recent plots list.
    """
    for s in ('sortby','what','job_type','starttime','endtime','regex','graph','group','image_width','image_height','image_dpi','period','use_period','title'):
      if s in kwargs and kwargs[s]:
        value = self._verify(s,urllib.unquote(kwargs[s]))
        #session['prodmon.plot.%s' % (s)]=value
        session['prodmon.plot.open'][session['prodmon.plot.current']][s]=value
    
    self._updateURLs(session)
    self.gui._saveSession(session)
    return self._state(session)
    
  def sessionFilters(self,session,*args,**kwargs):
    if 'name' in kwargs and kwargs['name']:
      name = urllib.unquote(kwargs['name'])
      if 'value' in kwargs and kwargs['value']:
        value = urllib.unquote(kwargs['value'])
        filters=[]
        if '|' in name and '|' in value:
          for n,v in zip(name.split('|'),value.split('|')):
            filters.append([n,v])
        else:
          filters.append([name,value])
        session['prodmon.plot.open'][session['prodmon.plot.current']]['filters']=filters
    if 'clear' in kwargs:
      session['prodmon.plot.open'][session['prodmon.plot.current']]['filters']=[]
    self._updateURLs(session)
    self.gui._saveSession(session)
    return self._state(session)
    
  def sessionSave(self,session,*args,**kwargs):
    #save the current plot
    if 'name' in kwargs and kwargs['name']:
      name = urllib.unquote(kwargs['name'])
      session['prodmon.plot.saved'][name]=deepcopy(session['prodmon.plot.open'][session['prodmon.plot.current']])
      
    self.gui._saveSession(session)  
    return self._state(session)
    
  def sessionEdit(self,session,*args,**kwargs):
    #rename a saved plot
    if 'newname' in kwargs and kwargs['newname']:
      if 'oldname' in kwargs and kwargs['oldname']:
        newname = urllib.unquote(kwargs['newname'])
        oldname = urllib.unquote(kwargs['oldname'])
        if oldname in session['prodmon.plot.saved']:
          session['prodmon.plot.saved'][newname]=deepcopy(session['prodmon.plot.saved'][oldname])
          del session['prodmon.plot.saved'][oldname]
    self.gui._saveSession(session)
    return self._state(session)       
   
  def sessionClose(self,session,*args,**kwargs):
    if 'index' in kwargs and kwargs['index'].isdigit():
      index = int(kwargs['index'])
    else:
      index = session['prodmon.plot.current']
    if index>=0 and index<len(session['prodmon.plot.open']):
      del session['prodmon.plot.open'][index]
    if len(session['prodmon.plot.open'])==0:
      session['prodmon.plot.open']=[deepcopy(ProdmonWorkspace.ProdmonPlot.plotdef)]
    session['prodmon.plot.current']=(session['prodmon.plot.current']+1)%len(session['prodmon.plot.open'])
    self._updateURLs(session)
    self.gui._saveSession(session)
    return self._state(session)
          
  def sessionDelete(self,session,*args,**kwargs):
    #delete a saved plot
    if 'name' in kwargs and kwargs['name']:
      name = urllib.unquote(kwargs['name'])
      if name in session['prodmon.plot.saved']:
        del session['prodmon.plot.saved'][name]
    self.gui._saveSession(session)
    return self._state(session)
    
  def sessionAdd(self,session,*args,**kwargs):
    #add a new graph window
    if 'copy' in kwargs:
      copy = kwargs['copy']
    else:
      copy=False
    index=-1
    if 'index' in kwargs and kwargs['index'].isdigit():
      index=int(kwargs['index'])
      if not (index>=0 and index<len(session['prodmon.plot.open'])):
        index=-1
    if copy:
      newplot = deepcopy(session['prodmon.plot.open'][session['prodmon.plot.current']])
    else:
      newplot = deepcopy(ProdmonWorkspace.ProdmonPlot.plotdef)
    if index==-1:
      session['prodmon.plot.open'].append(newplot)
      session['prodmon.plot.current']=len(session['prodmon.plot.open'])
    else:
      session['prodmon.plot.open'].insert(index,newplot)
      session['prodmon.plot.current']=index
    self.gui._saveSession(session)
    return self._state(session)
        
  def sessionLoad(self,session,*args,**kwargs):
    #load a saved session
    if 'new' in kwargs:
      new = kwargs['new']
    else:
      new = False
    if 'name' in kwargs and kwargs['name']:
      name = urllib.unquote(kwargs['name'])
      if name in session['prodmon.plot.saved']:
        newplot = deepcopy(session['prodmon.plot.saved'][name])
        if new:
          session['prodmon.plot.open'].append(newplot)
          session['prodmon.plot.current']=len(session['prodmon.plot.open'])
        else:
          session['prodmon.plot.open'][session['prodmon.plot.current']]=newplot
    self.gui._saveSession(session)
    return self._state(session)
      
  def sessionMove(self,session,*args,**kwargs):
    if 'index' in kwargs and kwargs['index'].isdigit():
      index = int(kwargs['index'])
      #if index>=0 and index<len(session['prodmon.plot.open']):
        #session['prodmon.plot.current']=index
      session['prodmon.plot.current']=index%len(session['prodmon.plot.open'])
    self.gui._saveSession(session)
    return self._state(session)
  

class ProdmonSource:
  url = 'http://dashb-cmspa.cern.ch/dashboard/request.py/PAquery?'
  plothook = "prodmon"
  params={}
  params['sortby']={'site':'Site','wf':'Workflow','dts':'Dataset','exit_code':'CMSSW Exit Code','prod_team':'Production Team','prod_agent':'ProdAgent'}
  params['what']={'evts_read_failed':'Events Read [Failed]','evts_read_success':'Events Read [Success]','evts_read_any':'Events Read','evts_written_failed':'Events Written [Failed]','evts_written_success':'Events written [Success]','evts_written_any':'Events Written','time_failed':'Time [Failed]','time_success':'Time [Success]','time_any':'Time','jobs_failed':'Jobs [Failed]','jobs_success':'Jobs [Success]','jobs_any':'Jobs','CMSSW_fail_codes':'CMSSW Failure Codes','time_by_span':'Approximate Batch Slot Usage','reco_rate':'Tier 0 RECO Rate','jobs_success_rates':'Job Success Rate','dts_timings':'Dataset Timings (Unnormalized)','jobs_failure_types':'Job Failure Types','wf_summary':'Workflow Summary Table'}
  params['plot'] = {'bar': 'Bar', 'pie':'Pie', 'cumulative':'Cumulative'}
  params['job_type'] = {'Merge': 'Merge', 'Any': 'Merge and Processing', 'Processing': 'Processing'}
  params['starttime']='Start Time'
  params['endtime']='End Time'
  filters={'site':'Site','wf':'Workflow','dts':'Dataset','exit_code':'CMSSW Exit Code','prod_team':'Production Team','prod_agent':'ProdAgent'}
  
  
  types={'pdf':'application/pdf','ps':'application/postscript','eps':'application/postscript','png':'image/png','svg':  'image/svg+xml','json':'text/plain'}
  sitelist = ('T0_CH_CERN', 'T1_CH_CERN', 'T1_DE_FZK', 'T1_ES_PIC', 'T1_FR_CCIN2P3', 'T1_IT_CNAF', 'T1_TW_ASGC', 'T1_UK_RAL', 'T1_US_FNAL',
         'T2_AT_Vienna', 'T2_BE_IIHE', 'T2_BE_UCL', 'T2_BR_SPRACE', 'T2_BR_UERJ', 'T2_CH_CAF', 'T2_CH_CSCS', 'T2_CN_Beijing',
          'T2_DE_DESY', 'T2_DE_RWTH', 'T2_EE_Estonia', 'T2_ES_CIEMAT', 'T2_ES_IFCA', 'T2_FI_HIP', 'T2_FR_CCIN2P3', 'T2_FR_GRIF_IRFU',
           'T2_FR_GRIF_LAL', 'T2_FR_GRIF_LLR', 'T2_FR_GRIF_LPNHE', 'T2_FR_IPHC', 'T2_HU_Budapest', 'T2_IN_TIFR', 'T2_IT_Bari',
            'T2_IT_Legnaro', 'T2_IT_Pisa', 'T2_IT_Rome', 'T2_KR_KNU', 'T2_PL_Warsaw', 'T2_PT_LIP_Coimbra', 'T2_PT_LIP_Lisbon',
             'T2_RU_IHEP', 'T2_RU_INR', 'T2_RU_ITEP', 'T2_RU_JINR', 'T2_RU_PNPI', 'T2_RU_RRC_KI', 'T2_RU_SINP', 'T2_TR_METU',
              'T2_TR_ULAKBIM', 'T2_TW_Taiwan', 'T2_UA_KIPT', 'T2_UK_London_Brunel', 'T2_UK_London_IC','T2_UK_London_Imperial', 'T2_UK_SGrid_Bristol','T2_UK_SouthGrid_Bristol',
               'T2_UK_SGrid_RALPP','T2_UK_SouthGrid_RALPP', 'T2_US_Caltech', 'T2_US_Florida', 'T2_US_MIT', 'T2_US_Nebraska', 'T2_US_Purdue', 'T2_US_UCSD',
                'T2_US_Wisconsin', 'T3_UK_London_QMUL', 'T3_UK_London_RHUL', 'T3_UK_London_UCL', 'T3_UK_ScotGrid_ECDF', 'T3_CH_PSI',
                 'T3_CN_PKU', 'T3_CO_Uniandes', 'T3_DE_Karlsruhe', 'T3_ES_Oviedo', 'T3_FR_IPNL', 'T3_GR_Demokritos', 'T3_GR_IASA',
                  'T3_GR_Ioannina', 'T3_IT_Napoli', 'T3_IT_Perugia', 'T3_IT_Trieste', 'T3_TW_NCU', 'T3_UK_SGrid_Oxford', 'T3_US_Colorado',
                   'T3_US_Cornell', 'T3_US_FNALLPC', 'T3_US_Kansas', 'T3_US_Minnesota', 'T3_US_Princeton', 'T3_US_Rice', 'T3_US_Rutgers',
                    'T3_US_TTU', 'T3_US_UCLA', 'T3_US_UCR', 'T3_US_UIowa', 'T3_US_UMD', 'T3_US_Vanderbilt')

  national={'CH':'Switzerland','DE':'Germany','ES':'Spain','IT':'Italy','TW':'Taiwan','UK':'United Kingdom','US':'United States','AT':'Austria','BE':'Belgium','BR':'Brazil','CN':'China','EE':'Estonia','FI':'Finland','HU':'Hungary','IN':'India','KR':'Korea','PK':'Pakistan','PL':'Poland','PT':'Portugal','RU':'Russia','TR':'Turkey','UA':'Ukraine','CO':'Colombia','GR':'Greece'}
  regional={}
  groups={}
  groups['site']={'null':'None','tier':'Tier','country':'Country','tiercountry':'Tier, Country','top5site':'Top 5 Sites'}
  groups['wf']={'null':'None','1':'Primary','2':'Secondary','3':'Tertiary','top5wf':'Top 5 Workflows'}
  groups['dts']={'null':'None','generator':'Generator/Dataset','campaign':'Production Campaign','conditions':'Conditions','tiers':'Data Tiers','merged':'Merge Status'}
  groups['exit_code']={'null':'None','10k':'10k'}
  groups['prod_team']={}
  groups['prod_agent']={}
  def __init__(self,server,*args):
    socket.setdefaulttimeout(60)
    self.cache = ProdmonCache()
    engine.on_stop_engine_list.append(self.cache.stop)
    self.cache.start()
  def plot(self,*path,**attrs):
    """
    Landing function for plotfairy requests.
    Format in *path should be what/sortby/plot/jobtype/datatype?starttime=XXX&endtime=...
    *path arguments are mandatory and this function should fail and return None,None if they are not found.
    **kwargs are all optional, defaults are provided in all cases.
    This function formats the query into a ProdmonQuery object, then passes it to GetPlotFromQuery.
    Defaults for missing keyword args are 48h plot, png image 800x600@72dpi (although these defaults are applied within ProdmonQuery)
    
    Also handles summary plots with paths of the form summary/{site,wf}/{plot,legend,map}
    An appropriate query object is created for these and then they are generated by the normal code path.
    """
      
    
    if len(path) == 3:
      if path[0] in ('summary') and path[1] in ('site','wf','team') and path[2] in ('plot','legend','map','plot+legend'): # Deprecated table output, now handled in state
        resulttype=path[2]

        if path[1] == 'site':
          query = ProdmonQuery('evts_read_any','site','pie','Any',type='png',width='400',height='300',group='top5site',period='48')
        elif path[1] == 'wf':
          query = ProdmonQuery('evts_read_any','wf','pie','Any',type='png',width='400',height='300',group='top5wf',period='48')
        elif path[1] == 'team':
          query =  ProdmonQuery('evts_read_any','prod_team','pie','Any',type='png',width='400',height='300',period='48')
        else:
          return None,None # Type of plot we don't understand
        return self._getPlotFromQuery(query,resulttype)
    
    elif len(path) == 5:
      query = ProdmonQuery(*path,**attrs)
      resulttype=path[4]
      return self._getPlotFromQuery(query,resulttype)
      
    elif len(path) >= 2:
      if path[0]=='autocomplete':
        ac_type = path[1]
        ac_text = '/'.join(path[2:])
        if ac_type in ('wf','dts','site','prod_team','prod_agent','exit_code'):
          return self._autoComplete(ac_type,ac_text)    
      
    
   
    return None,None # Fall through if the path length didn't match 3 or 5, or the summary format was inconsistent

  def _getPlotFromQuery(self,query,resulttype):
    """
    GetPlotFromQuery evaluates a query object, testing it for validity before using the cache access functions to return the data.
    """
    if not (resulttype in ('plot','legend','map','plot+legend','thumbnail') or (query['plot']=='data' and resulttype=='table')):
      print "PS: Unknown or inconsistent result type [%s]" % (resulttype)
      return None,None
    if not query.isValid():
      print "PS: Query invalid"
      return None,None
    
    plot = query['plot']
    if plot=='data':
      #return self._plotData(resulttype,query)
      return 'text/html',self._cacheFetch(query.cacheKey('PA_TABLE'),self._plotData,query)
    if not plot in ('pie','bar','cumulative'):
      print "PS: Unknown plot types [%s]" % (plot)
      return None,None
    
    plotdata = self._cacheFetchWithLock(query.cacheKey('PA_PLOT'),self._plotMaker,query)
    if resulttype=='plot':
      return ProdmonSource.types[query['type']],plotdata[0]
    elif resulttype=='legend':
      return ProdmonSource.types[query['type']],plotdata[1]
    elif resulttype=='map':
      return 'text/plain',plotdata[2]
    elif resulttype=='plot+legend':
      #return ProdmonSource.types[query['type']],stitchPlotAndLegend(plotdata[0],plotdata[1])
      return ProdmonSource.types[query['type']],plotdata[3]
    elif resulttype=='thumbnail':
      return ProdmonSource.types[query['type']],plotdata[4]
    else:
      pass # Already validated that we have one of the above
  
  def _plotMaker(self,query):
    """
    PlotMaker is a cache maker function which returns the appropriate plot,legend,map tuple if they aren't found in cache.
    """
    plot = query['plot']
    if plot=='pie':
      plotdata = self._plotPie(query)
    elif plot=='bar':
      plotdata = self._plotBar(query)
    elif plot=='cumulative':
      plotdata = self._plotCumulative(query)
    else:
      pass # Plot type validity has already been checked
    #if query['type']=='png':
    plotdata = plotdata[0],plotdata[1],plotdata[2],stitchPlotAndLegend(plotdata[0],plotdata[1]),makeThumbnail(plotdata[0])
    #else:
      #plotdata = plotdata[0],plotdata[1],plotdata[2],''
    return plotdata,1800,sum([len(d) for d in plotdata])
      
  def _plotData(self,query):
    """
    PlotData makes a formatted HTML table from a query request (where the 'what' is data)
    """
    qresult = self._cacheFetchWithLock(query.cacheKey('PA_QUERY',False),self._getResult,query)
    result = "<table>"
    if len(qresult)>0:
      fields = qresult[0].keys()
      result+="<tr>"
      for f in fields:
        result+="<td>%s</td>" % (f)
      result+="</tr>\n"
      for item in qresult:
        result+="<tr>"
        for f in fields:
          result+="<td>%s</td>" % (item[f])
        result+="</tr>\n"
    result+="</table>\n"    
    return result,1800,len(result)
      
    
  def summaryList(self,which,n=5):
    if which=='site':
      return self._cacheFetch("PA_SITETOP%s"%(n),self._getSiteTop,n)
    elif which=='wf':
      return self._cacheFetch("PA_WFTOP%s"%(n),self._getWFTop,n)
    return None

  def _getFilterGroup(self,query):
    filterer = lambda x: True
    if 'regex' in query and not query['regex']=='':
      filterer = self._regexFilter(query['regex'])
    
    group = lambda x: x
    if 'group' in query:
      if query['group'] in ProdmonSource.groups[query['sortby']]:
        if query['sortby']=='site':
          group = self._groupSite(query['group'])
        if query['sortby']=='wf':
          group = self._groupWF(query['group'])
        if query['sortby']=='dts':
          group = self._groupDTS(query['group'])
        if query['sortby']=='exit_code':
          group = self._groupCode(query['group'])
        if query['sortby']=='prod_agent':
          group = self._groupAgent(query['group'])
        if query['sortby']=='prod_team':
          group = self._groupTeam(query['group'])
    return filterer,group

  def _generateMapURL(self,query):
    """
    This is hook to provide deeper behaviour when a graph is clicked on. The default implementation just goes to a filtered version of the current plot, but it would be possible to add any sort of behaviour here.
    """
    base_query = {'what':query['what'],
                  'sortby':query['sortby'],
                  'job_type':query['job_type'],
                  'graph':query['plot'],
                  'name':'plot'
                 }
    if 'group' in query:
      base_query['group']=query['group']
    if 'period' in query:
      base_query['use_period']='true'
      base_query['period']=query['period']
    else:
      base_query['use_period']='false'
      base_query['starttime']=query['starttime']
      base_query['endtime']=query['endtime']
    base_query['filter_name']=query['sortby']
    return 'view?'+urllib.urlencode(base_query)+'&filter_value=$n'
                  
  
  def _plotPie(self,query):
        
    plotkey = query.cacheKey("PA_PLOT")
    result = self._cacheFetchWithLock(query.cacheKey('PA_QUERY',False),self._getResult,query)
    
    filterer,group = self._getFilterGroup(query)
    
    names=[]
    data={}
    
    # Only need to handle name,value for pie data, epoch not included
    for item in result:
      self._fixItem(query,item)
      name = item['NAME']
      if filterer(name):
        value=float(item['VALUE'])
        if value>0.:
          name = group(name)
          if not name in names:
            names.append(name)
          if not name in data:
            data[name]=value
          else:
            data[name]+=value
    
    if query['type']=='json':
      json = "[%r]" % (data)
      return json,json,json
    
    legend = self._getLegend(names)
    
    plot = PieChart(data,legend,{
            'imageType':query['type'],
            'height':query['height'],
            'width':query['width'],
            'dpi':query['dpi'],
            'title': query.plotTitle(),
            'starttime':query._stampToTime(query['starttime']),
            'endtime':query._stampToTime(query['endtime']),
            'integral':True,
            'mapName':'mainplot',
            'map_href':'#',
            #'map_onmouseover':"return GUI.Plugin.ProdmonGraphBase.handleMapMouse('%s','$n','$v')" % query['sortby'],
            'map_onclick':"return GUI.Plugin.ProdmonPlot.handleMapClick('%s')" % self._generateMapURL(query),
            'map_title':'$n: $v' # A note on tooltips. From Netscape 6, gecko based browsers ignore the alt-tag for tooltip and only use the title attribute. MSIE draws either, with alt taking precedence. HTML standard technically requires alt, but this will only increase map size. Including an onmouseover trigger suppresses either from displaying (this could be used for something more elaborate, eg YUI tooltips, but currently isn't).
            })
    imgdata,legenddata,mapdata = plot.make()
    return imgdata,legenddata,mapdata
  
  def _getSeries(self,result,interval):
    epochset=set()
    for item in result:
      if 'EPOCH' in item:
        epochset.add(int(item['EPOCH']))
      else:
        epochset.add(0)
        item['EPOCH']=0
    if len(epochset)>0:
      first = min(epochset)
      last = max(epochset)
      series = []
      now = first
      while (now<=last):
        series.append((now,now+interval))
        now += interval
      return series
    else:
      return None

  def _getLegend(self,names):
    colors = [ "#7D818C", "#007092", "#75B3CE", "#B40019", "#FFDC8C", "#F79025",
         "#BBBEC5", "#231F20", "#e66266", "#fff8a9", "#7bea81", "#8d4dff",
         "#ffbc71", "#a57e81", "#baceac", "#00ccff", "#ccffff", "#ff99cc",
         "#cc99ff", "#ffcc99", "#3366ff", "#33cccc", "#ffffff", "#bbbbbb",
         "#777777", "#333333" ]
    legend = []
    for i,n in enumerate(names):
      legend.append({'key':n,'name':n,'attrs':{'facecolor':colors[i % len(colors)]}})
    return legend
  
  def _fixItem(self,query,item):
    if query['what']=='jobs_success_rates':
      item['VALUE']=float(item['SUCCESS_JOBS'])/(float(item['FAILED_JOBS'])+float(item['SUCCESS_JOBS']))
    if query['what']=='dts_timings':
      item['VALUE']=float(item['AVERAGE_S_PER_EVT'])
    
  def _plotBar(self,query):

    result = self._cacheFetchWithLock(query.cacheKey('PA_QUERY',False),self._getResult,query)
    
    filterer,group = self._getFilterGroup(query)  
    
    #series - ((start,end),(start2,end2)...)
    #data - ({name:value,name2:value2},{...},...)
    span,interval = query.getInterval()
    series=self._getSeries(result,interval)
    
    if series:
    
      names=[]
      epochindex={}
      data=[]
      for i,s in enumerate(series):
        epochindex[s[0]]=i
        data.append({})
    
      for item in result:
        self._fixItem(query,item)
        name = item['NAME']
        if filterer(name):
          epoch = int(item['EPOCH'])
          value = float(item['VALUE'])
          if value>0.:
            name = group(name)
            if not name in names:
              names.append(name)
            if not name in data[epochindex[epoch]]:
              data[epochindex[epoch]][name]=value
            else:
              data[epochindex[epoch]][name]+=value
    
      #this is an ugly hack to ensure there is a data item known to the legend in each epoch
      #otherwise we get a daft blank graph
      if len(names)>=1:
        for d in data:
          if len(d)==0:
            d[names[0]]=0
    
    else:
      starttime = query._stampToTime(query['starttime'])
      endtime = query._stampToTime(query['endtime'])
      periods = int((endtime-starttime)/interval)
      series = [(starttime+n*interval,starttime+(n+1)*interval) for n in range(periods)]
      data=[{} for n in range(periods)]
      names=[]
    
    if query['type']=='json':
      json = "[%r,%r]" % (series,data)
      return json,json,json
    
    legend = self._getLegend(names)
    
    plot = TimeSeriesBarChart(span, series, data, legend, {
                  'imageType':query['type'],
                  'height':query['height'],
                  'width':query['width'],
                  'dpi':query['dpi'],
                  'title': query.plotTitle(),
                  'yAxisTitle': ProdmonSource.params['what'][query['what']],
                  'integral':True,
                  'mapName':'mainplot',
                  'map_href':'#',
                  'map_title':'$n: $v for $b-$e',
                  'map_onclick':"return GUI.Plugin.ProdmonPlot.handleMapClick('%s')" % self._generateMapURL(query),
                  #'map_onmouseover':"return GUI.Plugin.ProdmonGraphBase.handleMapMouse('%s','$n','$v')" % query['sortby']
                  })
    imgdata,legenddata,mapdata = plot.make()
    return imgdata,legenddata,mapdata
    
  
  def _plotCumulative(self,query):
    
    result = self._cacheFetchWithLock(query.cacheKey('PA_QUERY',False),self._getResult,query)
    
    filterer,group = self._getFilterGroup(query)  
    
    #series - ((start,end),(start2,end2)...)
    #data - ({name:value,name2:value2},{...},...)
    span,interval = query.getInterval()
    series=self._getSeries(result,interval)
    
    if series:  
      names=[]
      epochindex={}
      data=[]
      for i,s in enumerate(series):
        epochindex[s[0]]=i
        data.append({})
    
      for item in result:
        self._fixItem(query,item)
        name = item['NAME']
        if filterer(name):
          epoch = int(item['EPOCH'])
          value = float(item['VALUE'])
          if value>0.:
            name = group(name)
            if not name in names:
              names.append(name)
            if not name in data[epochindex[epoch]]:
              data[epochindex[epoch]][name]=value
            else:
              data[epochindex[epoch]][name]+=value
    
      #this is an ugly hack to ensure there is a data item known to the legend in each epoch
      #otherwise we get a daft blank graph
      #who knows why?
      if len(names)>=1:
        for d in data:
          if len(d)==0:
            d[names[0]]=0
    
    else:
      starttime = query._stampToTime(query['starttime'])
      endtime = query._stampToTime(query['endtime'])
      periods = int((endtime-starttime)/interval)
      series = [(starttime+n*interval,starttime+(n+1)*interval) for n in range(periods)]
      data=[{} for n in range(periods)]
      names=[]
    
    if query['type']=='json':
      json = "[%r,%r]" % (series,data)
      return json,json,json
    
    legend = self._getLegend(names)
    
    plot = CumulativeTimeSeriesChart(span, series, data, legend, {
                  'imageType':query['type'],
                  'height':query['height'],
                  'width':query['width'],
                  'dpi':query['dpi'],
                  'title': query.plotTitle(),
                  'yAxisTitle': ProdmonSource.params['what'][query['what']],
                  'integral':True,
                  'mapName':'mainplot',
                  'map_href':'#',
                  'map_title':'$n: $v',
                  'map_onclick':"return GUI.Plugin.ProdmonPlot.handleMapClick('%s')" % self._generateMapURL(query),
                  #'map_onmouseover':"return GUI.Plugin.ProdmonGraphBase.handleMapMouse('%s','$n','$v')" % query['sortby']
                  })
    imgdata,legenddata,mapdata = plot.make()
    return imgdata,legenddata,mapdata
    
    
  def _cacheFetchWithLock(self,cachekey,f_produce,*args,**kwargs):
    """
    Extended version of the _cacheFetch function that also locks the relevant key before fetching anything.
    """
    self.cache.keylock_acquire(cachekey)
    try:
      return self._cacheFetch(cachekey,f_produce,*args,**kwargs)
    finally:
      self.cache.keylock_release(cachekey)

  def _cacheFetch(self,cachekey,f_produce,*args,**kwargs):
    """
    Wrapper function to perform the fetch-from-cache-or-produce logic that was otherwise duplicated frequently.
    *args and **kwargs are passed to the f_produce function, which should be a callable that returns (data,lifetime,datasize)
    """
    if cachekey in self.cache:
      cachedata = self.cache[cachekey]
    else:
      cachedata,lifetime,size = f_produce(*args,**kwargs)
      if lifetime>0:
        self.cache.insert(cachekey,lifetime,size,cachedata)
    return cachedata
  
  def _getResult(self,query):
    """
    GetResult is a cachemaker function which actually gets data from the backend. HTTP errors are not actually handled here (logged and re-raised) since we don't want dead data to get cached - better to end up with HTTP 500 instead
    """
    urlquery = query.getURL()
    try:
      url = "%s%s" % (ProdmonSource.url,urlquery)
      data = urllib2.urlopen(url).read()
    #except IOError,e:
    #  print "PS: Error occurred during data fetching"
    #  print "PS: URL was %s" % (url)
    #  print "PS: Reason was %s" % (e.reason)
    #  raise # Re-raise the error
    except urllib2.HTTPError,e:
      print "PS: HTTPError during data fetching"
      print "PS: URL was %s" % (url)
      print "PS: Code was %s" % (e.code)
      raise
    except urllib2.URLError,e:
      print "PS: URLError during data fetching"
      print "PS: URL was %s" % (url)
      print "PS: Reason was %s" % (e.reason)
      raise
    if ("Exception" in data.split("\n",1)[0]):
      raise IOError, "Dashboard returned error %s in query %s" % (data.split("\n",1)[0],self.urlquery)
    result = ProdmonQueryResult(data)
    return result,1200,result.size()

  def _getErrorImage(self,error):
    """
    Produce a fake chart containing just an error message.
    """
    plot=ErrorChart(error,{imagetype:'png',height:640,width:480,dpi:72})
    return plot.draw().read()
      

  def getSiteList(self,sitefilter=None):
    """
    Get an optionally-filtered list of sites. Currently from a flat hardcode list.
    """
    return filter(sitefilter,ProdmonSource.sitelist)
  
  def _getWFList(self):
    """
    Get the list of all known workflows from the wf_summary table.
    Updated _getWFList for use with _fetchCache
    """
    query = ProdmonQuery('wf_summary','wf','data','Any')
    qresult = self._cacheFetchWithLock(query.cacheKey('PA_QUERY',False),self._getResult,query)
    result = []
    for item in qresult:
      result.append(copy(item['WORKFLOW'].encode('ascii')))
    size = 0
    for item in result:
      size += len(item)
    return result,1800,size
    
  def getAllWFSummary(self):
    return self._cacheFetch('PA_WFSUMMARY',self._getWFSummary)
    
  def getWFSummary(self,wf):
    table = self._cacheFetch('PA_WFSUMMARY',self._getWFSummary)
    return table[wf]
    
  def getStatsSummary(self):
    table = self._cacheFetch('PA_WFSUMMARY',self._getWFSummary)
    result = {}
    result['workflow_count']=len(table)
    result['total_unmerged']=0
    result['total_merged']=0
    time_per_event = []
    for t in table:
      result['total_unmerged']+=int(table[t]['UNMERGED_EVTS'])
      result['total_merged']+=int(table[t]['MERGED_EVTS'])
      time_per_event.append(float(table[t]['TIME_PER_EVENT']))
    result['avg_time_per_event']='%.2f'%(sum(time_per_event)/len(time_per_event))
    result['min_time_per_event']='%.2f'%min(time_per_event)
    result['max_time_per_event']='%.2f'%max(time_per_event)
    return result
  
  def _getWFSummary(self):
    """
    GetWFSummary fetches the special-case wf summary table, and parses it into a dictionary.
    """
    query = ProdmonQuery('wf_summary','wf','data','Any')
    qresult = self._cacheFetchWithLock(query.cacheKey('PA_QUERY',False),self._getResult,query)
    result = {}
    for item in qresult:
      itemresult={}
      wf = copy(item['WORKFLOW'])
      for attrib in item:
        itemresult[attrib] = copy(item[attrib])
      result[wf]=itemresult
    size = 0
    for wf in result:
      for attrib in result[wf]:
        size += len(result[wf][attrib])
    return result,1800,size
  
  def _getWFTop(self,count,sortkey='MERGED_EVTS'):
    wfsummary = self._cacheFetch('PA_WFSUMMARY',self._getWFSummary)
    
    wflist = wfsummary.keys()
    wflist.sort(key=lambda x: int(wfsummary[x][sortkey]))
    wflist.reverse()
    wflist = wflist[0:count]
    size = 0
    for wf in wflist:
      size+=len(wf)
    return wflist,1800,size
  
  def _getSiteTop(self,count):
    query = ProdmonQuery('evts_read_any','site','pie','Any')
    result = self._cacheFetchWithLock(query.cacheKey('PA_QUERY',False),self._getResult,query)
    
    sites = [(item['NAME'],int(item['VALUE'])) for item in result]
    sites.sort(key=lambda x: x[1])
    sites.reverse()
    
    sitelist = [s[0] for s in sites[0:count]]
    
    size=0
    for site in sitelist:
      size+=len(site)
    return sitelist,1800,size
  
  def getWFList(self,wffilter=None):
    if not wffilter:
      wffilter = lambda x: True
    wflist = self._cacheFetch("PA_WFLIST",self._getWFList)
    return filter(wffilter,wflist)

                
  def prepareSession(self,session):
    """
    Prepare when a new session hits the server. We don't handle sessions in the source, so do nothing.
    """
    pass
  
  def _regexFilter(self,regex):
    class _re_filter:
      def __init__(self,regex):
        try:
          self.regex = re.compile(regex)
        except re.error:
          self.regex = None
      def __call__(self,x):
        if self.regex:
          try:
            return self.regex.search(x)!=None
          except re.error:
            return False
        return False
    return _re_filter(regex)
          
    
  def _groupSite(self,mode):
    """
    GroupSite is a factory method that returns a callable() object which groups site names according to the selected mode.
    
    TODO: Regional doesn't work yet.
    TODO: Many of these functions should be delegated to sitedbsource.
    """
    
    class site_grouper:
      
      grouping={}
      grouping['tier']=lambda x: x.split('_')[0]
      grouping['country']=lambda x: self.national.get(x.split('_')[1],x.split('_')[1])
      grouping['tiercountry']=lambda x: "%s %s" % (x.split('_')[0], self.national.get(x.split('_')[1],x.split('_')[1]))
      def __init__(self,mode):
        self.function = self.grouping[mode]
      def __call__(self,site):
        try:
          return self.function(site)
        except IndexError:
          return site
    
    class top5_group_site:
      def __init__(self,toplist):
        self.toplist=toplist
      def __call__(self,site):
        if site in self.toplist:
          return site
        else:
          return 'Other'
    
    if mode in site_grouper.grouping:
      return site_grouper(mode)
    else:
      if mode=='top5site':
        return top5_group_site(self._cacheFetch("PA_SITETOP5",self._getSiteTop,5))
    return lambda x: x
      

  def _groupWF(self,mode):
    """
    GroupWF is a factory method that returns a callable for workflow grouping.
    """
    class count_group_wf:
      separator='-'
      def __init__(self,count):
        self.count = count
      def __call__(self,wf):
        if wf.count(self.separator)>=self.count:
          parts = wf.split(self.separator)
          return self.separator.join(parts[0:self.count])
        else:
          return wf
    class top5_group_wf:
      def __init__(self,toplist):
        self.toplist=toplist
      def __call__(self,wf):
        if wf in self.toplist:
          return wf
        else:
          return 'Other'
    
    if mode.isdigit():
      mode = int(mode)
      return count_group_wf(mode)
    else:
      if mode=='top5wf':
        return top5_group_wf(self._cacheFetch("PA_WFTOP5",self._getWFTop,5))
    return lambda x: x
    
  def _groupDTS(self,mode):
    class dts_grouper:
      def merged(x):
        if 'unmerged' in x:
          return 'UnMerged'
        else:
          return 'Merged'
      grouping={}
      grouping['generator']=lambda x: x.split('/')[1]
      grouping['campaign']=lambda x: x.split('/')[2].split('_')[0]
      grouping['conditions']=lambda x: x.split('/')[2].split('_')[1]
      grouping['tiers']=lambda x: x.split('/')[3]
      grouping['merged']=merged
      def __init__(self,mode):
        self.function = self.grouping[mode]
      
      def __call__(self,x):
        try:
          return self.function(x)
        except IndexError:
          return site      
    return dts_grouper(mode)
  
  def _groupCode(self,mode):
    if mode=='10k':
      def k10(x):
        if x.isdigit():
          return "%sxxxx" % ((int(x)%10000)/10000)
        return x
      return k10
    return lambda x:x
  
  def _groupAgent(self,mode):
    return lambda x:x
  
  def _groupTeam(self,mode):
    return lambda x:x
   
  def _autoComplete(self,mode,text):
    return 'text/plain','[]'

class ProdmonQuery:
  """
  Wrapper class for a query to the backend system. Turns it into a query string, adds timestamps if none exists and checks for validity.
  """
  timedelta = 3600 #all time variables are set to previous 30 minute interval, to make them cacheable (a request for 12:47:19 is converted to 12:30:00 etc)
  def __init__(self,what,sortby,plot,job_type,*args,**attrs):
    """
    Create a new Query from the path (as passed to ProdmonSource.plot) and attributes (from which we're only interested in time keys).
    Time variables are rounded to the nearest timedelta (30 minutes) - data isn't available at shorter timespans than this and otherwise no data is going to get cached.
    """
    self.query = {'what':what, 'sortby':sortby,'plot':plot,'job_type':job_type}
    #if no endtime is set, set it to now (with timedelta)
    if 'starttime' in attrs and 'endtime' in attrs:
      starttime = self._stampToTime(urllib.unquote(attrs['starttime']))
      endtime = self._stampToTime(urllib.unquote(attrs['endtime']))
      if not (starttime<endtime and endtime<time.time()):
        starttime = time.time()-172800
        endtime = time.time()
    elif 'period' in attrs:
      period = urllib.unquote(attrs['period'])
      if period.isdigit():
        period = int(period)
        starttime = time.time()-period*3600
        endtime = time.time()
      else:
        starttime = time.time()-172800
        endtime = time.time()
    else:
      starttime = time.time()-172800
      endtime = time.time()
    
    starttime = starttime - (starttime % ProdmonQuery.timedelta)
    endtime = endtime - (endtime % ProdmonQuery.timedelta)
    self.query['starttime']=self._timeToStamp(starttime)
    self.query['endtime']=self._timeToStamp(endtime)
      
    #these are known parameters which need to be considered in the plot cache key
    for s in ('filter_site','filter_wf','filter_dts','filter_prod_team','filter_prod_agent','filter_exit_code','regex','group','width','height','dpi','type','period'):
      if s in attrs and not (attrs[s]=='' or attrs[s]=='null'):
        self.query[s]=urllib.unquote(copy(attrs[s]))
    
    #set some defaults for optional image arguments
    if not 'height' in self.query:
      self.query['height']='600'
    if not 'width' in self.query:
      self.query['width']='800'
    if not 'dpi' in self.query:
      self.query['dpi']='72'
    if not 'type' in self.query:
      self.query['type']='png'
    else:
      if not self.query['type'] in ProdmonSource.types:
        self.query['type']='png'
    
  def getURL(self):
    """
    Return the query encoded to a url, ready to be appended to the source URL+?
    """
    #parameters understood by PAQuery
    #urlparam=('what','sortby','plot','job_type','starttime','endtime','span','site','wf','dts','prod_team')
    toencode={}
    #for up in urlparam:
      #if up in self.query:
        #toencode[up]=self.query[up]
    for q in self.query:
      if q in ProdmonSource.params:
        toencode[q]=self.query[q]
      if q.startswith('filter_'):
        if q[7:] in ProdmonSource.filters:
          toencode[q[7:]]=self.query[q]
    return urllib.urlencode(toencode)
    
  def isValid(self):
    """
    Test if the query is valid. Tests that all the mandatory parameters are filled (or the wf_summary special case), and that the start and end times are properly set and in the correct order.
    """
    
    matched = 0
    for key in self.query:
      if key in ProdmonSource.params:
        if self.query[key] in ProdmonSource.params[key]:
          matched += 1
    if matched==4 or (matched==2 and self.query['what']=='wf_summary' and self.query['plot']=='data'):
      if 'starttime' in self.query and 'endtime' in self.query:
        try:
          starttime = self._stampToTime(self.query['starttime'])
          endtime = self._stampToTime(self.query['endtime'])
        except ValueError:
          return False
        if starttime<=endtime and endtime<=time.time():
          return True
      else:
        return False
    else:
      return False
      
  def cacheKey(self,prefix='',full=True):
    """
    Return this query as a key to use for the backing cache.
    
    Two forms available - short key containing just the query parameters (full=False) for use in caching the query, and the long form (full=True) containing the image parameters and post-query options like regexes.
    """
    base = "%s|%s|%s|%s|%s|%s" % (self.query['what'],self.query['sortby'],self.query['plot'],self.query['job_type'],self.query['starttime'],self.query['endtime'])
    filters = ["|f:%s:%s"%(f[7:],self.query[f]) for f in self.query.keys() if f[0:6]=='filter']
    base+='|'.join(filters)
    image=""
    if 'width' in self.query and 'height' in self.query and 'dpi' in self.query and 'type' in self.query:
      image = "|%s|%sx%sx%s" % (self['type'],self['width'],self['height'],self['dpi'])
    extra = ""
    if 'regex' in self.query:
      extra += "|regex:%s" % (self['regex'])
    if 'group' in self.query:
      extra += "|group:%s" % (self['group'])
    if full:
      return "%s@%s%s%s" % (prefix,base,image,extra)
    else:
      return "%s@%s" % (prefix,base)
  
  
  def _timeToStamp(self,time):
    """
    Convert a seconds since epoch time number into the query timestamp format.
    TODO: check timezone behaviour of this
    """
    dt = datetime.datetime.fromtimestamp(time)
    return "%s-%s-%s %s:%s:%s" % (dt.year,dt.month,dt.day,dt.hour,dt.minute,dt.second)
  def _stampToTime(self,stamp):
    """
    Convert a query timestamp into seconds since the epoch.
    TODO: check timezone behaviour of this
    """
    ts = time.mktime(time.strptime(stamp,"%Y-%m-%d %H:%M:%S"))
    return ts
  def __getitem__(self,key):
    """
    Direct access to the query keys. Don't you just love python operator overriding?
    """
    return self.query[key]
  
  def plotTitle(self):
    """
    Get a formatted plot title based on this query.
    """
    start = "%s (%s jobs)" % (ProdmonSource.params['what'][self['what']],ProdmonSource.params['job_type'][self['job_type']])
    sort = "sorted by %s" % (ProdmonSource.params['sortby'][self['sortby']])
    group = ""
    if 'group' in self.query and self.query['group'] in ProdmonSource.groups[self['sortby']]:
      group = "grouped by %s" % (ProdmonSource.groups[self['sortby']][self['group']])
    regex=""
    if 'regex' in self.query and not (self['regex']=='' or self['regex']=='null'):
      regex = "matching %s" % self['regex']
    
    filters = ["%s~%s"%(f[7:],self.query[f]) for f in self.query.keys() if f[0:6]=='filter']  
    return " ".join([start,sort,group,regex]+filters)
  
  def getInterval(self):
    """
    Get the time binning interval for this data. This should probably be done by examination of the data,
    but as it currently stands short of compiling a histogram this is impossible and it will instead be
    inferred on the basis of 1-day intervals for >168 hours and 1-hour intervals otherwise
    """
    start = self._stampToTime(self['starttime'])
    end = self._stampToTime(self['endtime'])
    if end-start >= 604800:
      return 'day',86400
    else:
      return 'hour',3600
    
  
  
  def __iter__(self):
    return self.query.__iter__()

  
class ProdmonQueryResult:
  """
  Parses the raw data from an xml request (using DOM) and provides an iterator over the items it returns.
  """
  def __init__(self,data):
    """
    Create a new Query result from raw XML data
    """
    dom = parseString(data)
    self.items=[]
    for item in dom.getElementsByTagName("item"):
      result={}
      for c in item.childNodes:
        tagname = copy(c.tagName)
        if c.firstChild:
          child = copy(c.firstChild.nodeValue.encode('ascii'))
          result[tagname]=child
        else:
          result[tagname]=''
      self.items.append(result)
    dom.unlink()
    
  def size(self):
    size=0
    for item in self.items:
      for k,v in item.iteritems():
        size+=len(k)
        size+=len(v)
    return size
  
  def __getitem__(self,i):
    """
    Get an indexed item, returning a dictionary of the named sub-keys and their values.
    """
    return self.items[i]
  def __iter__(self):
    """
    Return a QueryIterator over this result.
    """
    return self.items.__iter__()
  def __len__(self):
    """
    Get the number of items this result contains.
    """
    return len(self.items)
    
  def __str__(self):
    return str(self.items)
    
  def __repr__(self):
    return repr(self.items)
        
  
class ProdmonCache(Thread):
  """
  Thread-watched cache of data for prodmon (or anything else)
  All non _-prefixed methods should be thread-safe for external calling
  """
  
  #synchronized function decorator. use?
  #@synchronized(self.lock)
  def synchronized(lock):
    def wrap(f):
      def inner_wrap(*args,**kwargs):
        lock.acquire()
        try:
          return f(*args,**kwargs)
        finally:
          lock.release()
      return inner_wrap
    return wrap
  
  def __init__(self,interval=10,cachelimit=10000000):
    """
    Create a new ProdmonCache. Note that it is not started until the .start() method is called.
    
    @param interval: interval in seconds between the main loop running (30)
    @param cachesize: maximum size in bytes before old entries start getting culled. Note that this is done *imprecisely* and will not be strictly adhered to! (10MB)
    """
    Thread.__init__(self,name="ProdmonCache")
    self.lock = Lock()
    self.stopme = False
    self.cachesize = 0
    self.cache = {} # objkey: tuple (expiry, size, data)
    self.interval = interval
    self.cachelimit = cachelimit
    self.cachelocks = {} #objkey: tuple(expiry,data)
    
  def stop(self):
    """
    When called, causes the thread to exit, the next time the main loop runs.
    """
    self.lock.acquire()
    #print "PC: setting exit flag"
    self.stopme = True
    self.lock.release()
    
  def run(self):
    while True:
      self.lock.acquire()
      #print "PC: Main loop runs. %s items in cache using %s bytes." % (len(self.cache),self.cachesize)
      if self.stopme:
        #print "PC: Main loop exiting"
        return
      self._clean_expired()
      self.lock.release()
      time.sleep(self.interval)
  def _expiry_dict(self,cachedict):
    """
    Return a dictionary of of {expiry: [objkeys]}. Intended for internal use so no locking is performed!
    """
    result = {}
    for objkey in cachedict:
      expiry = self.cache[objkey][0]
      if expiry in result:
        result[expiry].append(objkey)
      else:
        result[expiry]=[]
        result[expiry].append(objkey)
    return result
  
#  def _size_dict(self):
#    """
#    Return a dictionary of of {datasize: objkey}. Intended for internal use so no locking is performed!
#    """
#    result = {}
#    for objkey in self.cache.keys():
#      size = self.cache[objkey][1]
#      if size in result:
#        result[size].append(objkey)
#      else:
#        result[size]=[]
#        result[size].append(objkey)
#    return result
    
  def _clean_expired(self):
    """
    Iterate through the cache, removing any keys for which the current time is later than the expiry time.
    Assumes that it will only be called from a function that has already acquired a lock.
    """
    
    now = time.time()
    exp_lockkeys=[]
    for lockkey in self.cachelocks:
      exp = self.cachelocks[lockkey][0]
      if exp<now:
        exp_lockkeys.append(lockkey)
    for l in exp_lockkeys:
      del self.cachelocks[l] # Deleting a Lock appears to release it. Is this always true?
        
    exp_keys=[]
    for objkey in self.cache:
      exp = self.cache[objkey][0]
      if exp<now:
        exp_keys.append(objkey)
    
    for k in exp_keys:
      self._remove(k)
  
  def keylock_acquire(self,key,expiry=60):
    lock=None
    self.lock.acquire()
    try:
      if key in self.cachelocks:
        lock=self.cachelocks[key][1]
      else:
        lock=Lock()
        self.cachelocks[key]=(expiry+time.time(),lock)
    finally:
      self.lock.release()
    #release global lock, acquire local lock
    lock.acquire()
  
  def keylock_release(self,key):
    lock=None
    self.lock.acquire()
    try:
      if key in self.cachelocks:
        lock=self.cachelocks[key][1]
    finally:
      self.lock.release()
    #release global lock, release local lock
    if lock:
      lock.release()

  def keylock_get(self,key):
    lock=None
    self.lock.acquire()
    try:
      if key in self.cachelocks:
        lock=self.cachelocks[key][1]
    finally:
      self.lock.release()
    #release global lock, release local lock
    return lock
      
  
  def _clean_volume(self,volume):
    """
    Iterate through the cache, removing keys in expiry order to try and recover volume space (or empty the cache completely).
    Assumes that it will only be called from a function that has already acquired a lock.
    @param volume: target volume to remove. 
    """
    #"PC: Cleaning volume [%s]" % (volume)
    cleaned = 0
    expiry_dict = self._expiry_dict(self.cache)
    expiry_list = sorted(expiry_dict.keys())
    while cleaned<=volume and len(expiry_list)>=0:
      expiry_time = expiry_dict[expiry_list.pop(0)]
      if type(expiry_time)==type([]):
        expiry_time=expiry_time[0]
      for key in expiry_dict[expiry_time]:
        cleaned += self._remove(key)
        if cleaned>=volume:
          break
      #print "PC: Deleted key [%s], recovered total [%s]" % (expiry_key,cleaned)
    
  
  def _insert(self,key,lifetime,size,data,withlock):
    #print "PC: Inserting key [%s], lifetime: %s, size: %s" % (key,lifetime,size)
    if self._exists(key):
      self._remove(key)
    self.cache[key]=(lifetime+time.time(),size,data)
    self.cachesize+=size
    if self.cachesize>=self.cachelimit:
      #print "PC: New insertion exceeds cache limit [%s/%s]" % (self.cachesize,self.cachelimit)
      self._clean_volume(self.cachesize-self.cachelimit)
    #print "PC: Insertion complete"
  
  def insert(self,key,lifetime,size,data,withlock=False):
    """
    Insert a new key into the cache, with associated expiry lifetime, size and data. Triggers volume cleaning if the new cache size exceeds the maximum.
    @param key: key for the new data - should be a string
    @param lifetime: desired cache lifetime of this object in seconds
    @param size: amount of space this object requires (since we can't trivially find this at insertion time, but at download or image creation time this may be better known)
    @param data: free data object. Preferably doesn't refer to other objects in the cache, or size calculations will become impossible.
    @param withlock: if you want a lock to be associated with this object
    """
    
    self.lock.acquire()
    try:
      self._insert(key,lifetime,size,data,withlock)
    finally:
      self.lock.release()
    
  
  def _remove(self,key):
    #print "PC: Removing key [%s]" % (key)
    size=-1
    if self._exists(key):
      size = self.cache[key][1]
      self.cachesize-=size
      del self.cache[key]
    return size
  
  def remove(self,key):
    """
    Remove a key from the cache, returning the freed space. Returns -1 if the key wasn't found (or if you set the key size to -1, which would be stupid.
    @param key: cache key to remove
    """
    
    self.lock.acquire()
    try:
      size = self._remove(key)
    finally:
      self.lock.release()
    return size
  
  def retrieve(self,key,default=None):
    """
    Attempts to retrieve the data associated with a key from the cache. Returns default (None) if not found.
    @param key: cache key to retrieve data for
    @param default: what to return if the key is not in the cache (None)
    """
    #print "PC: Getting key [%s]" % (key)
    self.lock.acquire()
    try:
      if self._exists(key):
        result = self.cache[key][2]
      else:
        #print "PC: Key [%s] not in cache" % (key)
        result= default
    finally:
      self.lock.release()
    return result
  
  def _exists(self,key):
    return key in self.cache
  
  def exists(self,key):
    result=False
    self.lock.acquire()
    try:
      result = self._exists(key)
    finally:
      self.lock.release()
    #print "PC: Testing key existence [%s], %s" % (key,result)
    return result
    
  def _expire(self,key):
    """
    Symonym for remove
    """
    return self.remove(key)
  
  def cachesize(self):
    """
    Get the current "size" of the cache
    """
    self.lock.acquire()
    try:
      result =  self.cachesize
    finally:
      self.lock.release()
    return result
    
  def cachelimit(self):
    """
    Get the sizelimit of the cache
    """
    self.lock.acquire()
    try:
      result= self.cachelimit
    finally:
      self.lock.release()
    return result
    
  def itemcount(self):
    """
    Get the number of items in the cache
    """
    self.lock.acquire()
    try:
      result= len(self.cache)
    finally:
      self.lock.release()
    return result
  
  def keylist(self):
    """
    Get a shallow-copied list of the keys in the cache
    """
    self.lock.acquire()
    try:
      result = self.cache.keys()
    finally:
      self.lock.release()
    return result
    
  def __getitem__(self,key):
    return self.retrieve(key)
  
  def __delitem__(self,key):
    return self.remove(key)
  
  def __len__(self):
    return self.itemcount()
  
  def __iter__(self):
    return self.keylist().__iter__()
  
  def __contains__(self,key):
    return self.exists(key)
    
  def plot(self,*args,**kwargs):
    if args[0]=='list':
      result="<table>\n"
      result+="<tr><td>Key</td><td>Expiry</td><td>Remaining</td><td>Size</td><td>Data</td></tr>\n"
      for key in self.cache:
        result+="<tr>%s</tr>\n"%''.join(['<td>%s</td>'%i for i in [key,self.cache[key][0],self.cache[key][0]-time.time(),self.cache[key][1],'<a href="/overview/cache/data/%s">data</a>'%(key)]])
      result+="</table>"
      return 'text/html',result
    if args[0]=='data':
      key = args[1]
      data = self.cache[key][2]
      if type(data)==type(()):
        result="<html><body>"
        for i in range(len(data)):
          result+="<img src='/overview/cache/item/%s/%s'><hr>"%(key,i)
        result+="</body></html>"
        return 'text/html',result
      else:
        return 'text/plain',repr(data)
    if args[0]=='item':
      key = args[1]
      item = int(args[2])
      return 'image/png',self.cache[key][2][item]
      
      
  
#if __name__ == "__main__":
#  pc = ProdmonCache(10,100000)
#  for i in range(1000):
#    pc.insert(str(i),random.randint(1,300),random.randint(1,100),None)
#  pc.start()
